(* Lexer.lex
 *
 * Neil Wilkie 2004-05
 * Edinburgh University Informatics UG4 Project
 *
 * 'Random Testing of ML Programs'
 *
 * This is the lexer for the lexing and parsing of axioms to a form by which
 * they can be represented in the program. It is based upon a lexer for
 * 'Micro-ML' from a UG3 Functional Programming and Specification
 * practical by Don Sannella.
 *)

{
 open Lexing Parser;

 exception LexicalError of string * int * int (* (message, loc1, loc2) *)

 fun lexerError lexbuf s = 
     raise LexicalError (s, getLexemeStart lexbuf, getLexemeEnd lexbuf);

 val commentStart = ref 0;  (* Start of outermost comment being scanned *)
 
 fun commentNotClosed lexbuf =
     raise LexicalError ("Comment not terminated", 
                         !commentStart, getLexemeEnd lexbuf);
     
 val commentDepth = ref 0;  (* Current comment nesting *)

 (* Scan keywords as identifiers and use this function to distinguish them. *)

 fun keyword s =
     case s of
         "true"         => TRUE
       | "false"        => FALSE
       | "andalso"      => ANDALSO
       | "orelse"       => ORELSE
       | "and"          => AND
       | "or"           => OR
       | "forall"       => FORALL
       | "exists"       => EXISTS
       | "int"          => INTT
       | "real"         => REAL
       | "word"         => WORD
       | "char"         => CHAR
       | "string"       => STRING
       | "mod"          => MOD
       | _              => NAME s;

 }

rule Token = parse
    [` ` `\t` `\n` `\r`]     { Token lexbuf }
  | `~`?[`0`-`9`]+      { case Int.fromString (getLexeme lexbuf) of
                               NONE   => lexerError lexbuf "internal error"
                             | SOME i => INT i
                        }  
  | [`a`-`z``A`-`Z`][`a`-`z``A`-`Z``0`-`9`]*
                        { keyword (getLexeme lexbuf) }
  | "(*"                { commentStart := getLexemeStart lexbuf;
                          commentDepth := 1; 
                          SkipComment lexbuf; Token lexbuf }
  | "=>"                { DOUBLEARROW }
  | `=`                 { EQ }
  | `+`                 { PLUS }                     
  | `-`                 { MINUS }                     
  | `*`                 { TIMES }                     
  | `/`                 { DIV }                     
  | `|`                 { BAR }                     
  | `(`                 { LPAR }
  | `)`                 { RPAR }
  | `>`                 { GRTRTHN }
  | `<`                 { LSSTHN }
  | "==>"               { IMPLIES }
  | `~`                 { NOT }
  | "."                 { DOT }
  | ":"                 { COLON }
  | ","                 { COMMA }
  | "#"                 { HASH }
  | "\""                { QUOTE }
  | eof                 { EOF }
  | _                   { lexerError lexbuf "Illegal symbol in input" }

and SkipComment = parse
    "*)"                { commentDepth := !commentDepth - 1;  
                          if !commentDepth = 0 then ()
                          else SkipComment lexbuf 
                        } 
   | "(*"               { commentDepth := !commentDepth + 1; 
                          SkipComment lexbuf }
   | (eof | `\^Z`)      { commentNotClosed lexbuf }
   | _                  { SkipComment lexbuf }
;
