(* qecheck.sml
 *
 * Neil Wilkie 2004-05
 * Edinburgh University Informatics UG4 Project
 *
 * 'Random Testing of ML Programs'
 *
 * This is the main code for QuickECheck. All required functions are
 * loaded in here. It is this file which is required to be loaded to
 * use QuickECheck.
 *)

(* Load in some libraries which we are going to need later on. *)

load "Int";
load "Bool";
load "String";
load "Real";
load "Word";
load "Char";
open TextIO;


(* We need to load in the original QCheck stuff. Because we're running
 * mosml with '-I qcheck' we need not worry about directories.
 *)

load "qcheck";
use "qcheck.sml";
open QCheck infix ==>;
open QCheck infix implies;


(* Load the parser and associated things. Because we're running mosml
 * with '-I parser' we need not worry about directories.
 *)

load "Data";
load "Lexer";
load "Parser";
load "Main";
use "Data.sml";
use "Lexer.sml";
use "Parser.sml";
use "Main.sml";


(* Define some exceptions which might need to be called if there is
 * an error.
 *)

exception Not_Useful_Test;
exception Too_Many_Vars;
exception Existing_Exists;
exception Program_Error;
 

(* Using the type information from the axiom we can create an
 * appropriate generator and toString function.
 *)

fun getVars(Disjunction(f0, f1), vars) = getVars(f1, getVars(f0, vars))
  | getVars(Conjunction(f0, f1), vars) = getVars(f1, getVars(f0, vars))
  | getVars(Implies(f0, f1), vars)     = getVars(f1, getVars(f0, vars))
  | getVars(Negation(f), vars)         = getVars(f, vars)
  | getVars(Forall(v, t, f), vars)     = getVars(f, vars@[(v, t)])
  | getVars(Exists(v, t, f), vars)     = getVars(f, vars@[(v, t)])
  | getVars(Atomic(a), vars)           = vars

and makeGen([(v, t)]) =
      getGen(t) ^ ", SOME " ^ getToString(t) 
  | makeGen([(v0, t0), (v1, t1)]) =
      "Gen.zip(" ^ getGen(t0) ^ ", " ^ getGen(t1) ^
      "), SOME (fn (t0,t1) => " ^ getToString(t0) ^
      " t0 ^ \", \" ^ " ^ getToString(t1) ^ " t1)"
  | makeGen([(v0, t0), (v1, t1), (v2, t2)]) =
      "Gen.zip3(" ^ getGen(t0) ^ ", " ^ getGen(t1) ^
      ", " ^ getGen(t2) ^ "), SOME (fn (t0,t1,t2) => " ^
      getToString(t0) ^ " t0 ^ \", \" ^ " ^ getToString(t1) ^
      " t1 ^ \", \" ^ " ^ getToString(t2) ^ " t2)"
  | makeGen([(v0, t0), (v1, t1), (v2, t2), (v3, t3)]) =
      "Gen.zip4(" ^ getGen(t0) ^ ", " ^ getGen(t1) ^ ", " ^
      getGen(t2) ^ ", " ^ getGen(t3) ^
      "), SOME (fn (t0,t1,t2) => " ^ getToString(t0) ^
      " t0 ^ \", \" ^ " ^ getToString(t1) ^ " t1 ^ \", \" ^ " ^
      getToString(t2) ^ " t2 ^ \", \" ^ " ^ getToString(t3) ^ " t3)"
  | makeGen(vars) = raise Too_Many_Vars

and getGen(Intt)   = "Gen.Int.int"
  | getGen(Real)   = "Gen.Real.real"
  | getGen(Word)   = "Gen.Word.word"
  | getGen(Char)   = "Gen.char"
  | getGen(String) =
      "Gen.string(Gen.range(5, 100), Gen.char)"
      (*"Gen.string(Gen.range(5,20),Gen.charFrom " ^
      "\"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz\")"*)

and getToString(Intt)   = "Int.toString"
  | getToString(Real)   = "Real.toString"
  | getToString(Word)   = "Word.toString"
  | getToString(Char)   = "Char.toString"
  | getToString(String) = "String.toString"


(* Work through the algorithm for manipulating the formula. First
 * simplify it by removing implications, then push the negations
 * inwards, and finally pull the qunatifiers to the outside.
 *)

fun simplify(Disjunction(f0, f1)) =
      Disjunction(simplify(f0), simplify(f1))
  | simplify(Conjunction(f0, f1)) =
      Conjunction(simplify(f0), simplify(f1))
  | simplify(Implies(f0, f1)) =
      Disjunction(Negation(simplify(f0)), simplify(f1))
  | simplify(Negation(Implies(f0, f1))) =
      Conjunction(simplify(f0), Negation(simplify(f1)))
  | simplify(Negation(f)) =
      Negation(simplify(f))
  | simplify(Forall(v, t, f)) =
      Forall(v, t, simplify(f))
  | simplify(Exists(v, t, f)) =
      Exists(v, t, simplify(f))
  | simplify(Atomic(a)) =
      Atomic(a)

and isSimplified(Disjunction(f0, f1)) =
      isSimplified(f0) andalso isSimplified(f1)
  | isSimplified(Conjunction(f0, f1)) =
      isSimplified(f0) andalso isSimplified(f1)
  | isSimplified(Implies(f0, f1)) =
      false
  | isSimplified(Negation(f)) =
      isSimplified(f)
  | isSimplified(Forall(v, t, f)) =
      isSimplified(f)
  | isSimplified(Exists(v, t, f)) =
      isSimplified(f)
  | isSimplified(Atomic(a)) =
      true

and keepSimplifying(f, finished) =
      if finished
      then f
      else keepSimplifying(simplify(f), isSimplified(simplify(f)))

and moveNegations(Disjunction(f0, f1)) = 
      Disjunction(moveNegations(f0), moveNegations(f1))
  | moveNegations(Conjunction(f0, f1)) =
      Conjunction(moveNegations(f0), moveNegations(f1))
  | moveNegations(Implies(f0, f1)) =
      Implies(moveNegations(f0), moveNegations(f1))
  | moveNegations(Negation(Negation(f))) =
      moveNegations(f)
  | moveNegations(Negation(Disjunction(f0, f1))) =
      Conjunction(Negation(moveNegations(f0)),
                  Negation(moveNegations(f1)))
  | moveNegations(Negation(Conjunction(f0, f1))) =
      Disjunction(Negation(moveNegations(f0)),
                  Negation(moveNegations(f1)))
  | moveNegations(Negation(Exists(v, t, f))) =
      Forall(v, t, Negation(moveNegations(f)))
  | moveNegations(Negation(f)) =
      Negation(moveNegations(f))
  | moveNegations(Forall(v, t, f)) =
      Forall(v, t, moveNegations(f))
  | moveNegations(Exists(v, t, f)) =
      Exists(v, t, moveNegations(f))
  | moveNegations(Atomic(a)) =
      Atomic(a)

and negationsMoved(Disjunction(f0, f1)) = 
      negationsMoved(f0) andalso negationsMoved(f1)
  | negationsMoved(Conjunction(f0, f1)) =
      negationsMoved(f0) andalso negationsMoved(f1)
  | negationsMoved(Implies(f0, f1)) =
      negationsMoved(f0) andalso negationsMoved(f1)
  | negationsMoved(Negation(Negation(f))) =
      false
  | negationsMoved(Negation(Disjunction(f0, f1))) =
      false
  | negationsMoved(Negation(Conjunction(f0, f1))) =
      false
  | negationsMoved(Negation(Exists(v, t, f))) =
      false
  | negationsMoved(Negation(f)) =
      negationsMoved(f)
  | negationsMoved(Forall(v, t, f)) =
      negationsMoved(f)
  | negationsMoved(Exists(v, t, f)) =
      negationsMoved(f)
  | negationsMoved(Atomic(a)) =
      true

and keepMovingNegations(f, finished) =
      if finished
      then f
      else keepMovingNegations(moveNegations(f),
			       negationsMoved(moveNegations(f)))

and moveQuantifiers(Disjunction(Forall(v, t, f0), f1)) =
      Forall(v, t, moveQuantifiers(Disjunction(f0, f1)))
  | moveQuantifiers(Disjunction(Exists(v, t, f0), f1)) =
      Exists(v, t, moveQuantifiers(Disjunction(f0, f1)))
  | moveQuantifiers(Disjunction(f0, Forall(v, t, f1))) =
      Forall(v, t, moveQuantifiers(Disjunction(f0, f1)))
  | moveQuantifiers(Disjunction(f0, Exists(v, t, f1))) =
      Exists(v, t, moveQuantifiers(Disjunction(f0, f1)))
  | moveQuantifiers(Disjunction(f0, f1)) = 
      Disjunction(moveQuantifiers(f0), moveQuantifiers(f1))
  | moveQuantifiers(Conjunction(Forall(v, t, f0), f1)) =
      Forall(v, t, moveQuantifiers(Conjunction(f0, f1)))
  | moveQuantifiers(Conjunction(Exists(v, t, f0), f1)) =
      Exists(v, t, moveQuantifiers(Conjunction(f0, f1)))
  | moveQuantifiers(Conjunction(f0, Forall(v, t, f1))) =
      Forall(v, t, moveQuantifiers(Conjunction(f0, f1)))
  | moveQuantifiers(Conjunction(f0, Exists(v, t, f1))) =
      Exists(v, t, moveQuantifiers(Conjunction(f0, f1)))
  | moveQuantifiers(Conjunction(f0, f1)) =
      Conjunction(moveQuantifiers(f0), moveQuantifiers(f1))
  | moveQuantifiers(Implies(f0, f1)) =
      Implies(moveQuantifiers(f0), moveQuantifiers(f1))
  | moveQuantifiers(Negation(f)) =
      Negation(moveQuantifiers(f))
  | moveQuantifiers(Forall(v, t, f)) =
      Forall(v, t, moveQuantifiers(f))
  | moveQuantifiers(Exists(v, t, f)) =
      moveQuantifiers(Negation(Forall(v, t, Negation(moveQuantifiers(f)))))
  | moveQuantifiers(Atomic(a)) =
      Atomic(a)

and quantifiersMoved(Disjunction(Forall(v, t, f0), f1)) =
      false
  | quantifiersMoved(Disjunction(Exists(v, t, f0), f1)) =
      false
  | quantifiersMoved(Disjunction(f0, Forall(v, t, f1))) =
      false
  | quantifiersMoved(Disjunction(f0, Exists(v, t, f1))) =
      false
  |quantifiersMoved(Disjunction(f0, f1)) = 
      quantifiersMoved(f0) andalso quantifiersMoved(f1)
  | quantifiersMoved(Conjunction(Forall(v, t, f0), f1)) =
      false
  | quantifiersMoved(Conjunction(Exists(v, t, f0), f1)) =
      false
  | quantifiersMoved(Conjunction(f0, Forall(v, t, f1))) =
      false
  | quantifiersMoved(Conjunction(f0, Exists(v, t, f1))) =
      false
  | quantifiersMoved(Conjunction(f0, f1)) = 
      quantifiersMoved(f0) andalso quantifiersMoved(f1)
  | quantifiersMoved(Implies(f0, f1)) =
      quantifiersMoved(f0) andalso quantifiersMoved(f1)
  | quantifiersMoved(Negation(f)) =
      quantifiersMoved(f)
  | quantifiersMoved(Forall(v, t, f)) =
      quantifiersMoved(f)
  | quantifiersMoved(Exists(v, t, f)) =
      quantifiersMoved(f)
  | quantifiersMoved(Atomic(a)) =
      true

and keepMovingQuantifiers(f, finished) =
      if finished
      then f
      else keepMovingQuantifiers(moveQuantifiers(f),
				 quantifiersMoved(moveQuantifiers(f)))

and checkQuantifiers(Disjunction(f0, f1)) =
      checkQuantifiers(f0) andalso checkQuantifiers(f1)
  | checkQuantifiers(Conjunction(f0, f1)) =
      checkQuantifiers(f0) andalso checkQuantifiers(f1)
  | checkQuantifiers(Implies(f0, f1)) =
      checkQuantifiers(f0) andalso checkQuantifiers(f1)
  | checkQuantifiers(Negation(f)) =
      checkQuantifiers(f)
  | checkQuantifiers(Forall(v, t, f)) =
      checkQuantifiers(f)
  | checkQuantifiers(Exists(v, t, f)) =
      false
  | checkQuantifiers(Atomic(a)) =
      true

and fullTransform(f) =
      keepMovingQuantifiers(keepMovingNegations(
						      keepSimplifying(f, false),
						      false), false)
      
and simpleTransform(f) =
      keepMovingQuantifiers(keepMovingNegations(f, false), false);


   


(* Create the function required to preform the test in QCheck by
 * unparsing the formula into the correct form.
 *)

fun create(Disjunction(f0, f1)) =
      create(f0) ^ " orelse " ^ create(f1)
  | create(Conjunction(f0, f1)) =
      create(f0) ^ " andalso " ^ create(f1)
  | create(Implies(Forall(v, t, f0), f1)) =
      raise Not_Useful_Test
  | create(Implies(f0, f1)) =
      "not (" ^ create(f0) ^ ") orelse " ^ create(f1)
  | create(Negation(Forall(v, t, f))) =
      raise Not_Useful_Test
  | create(Negation(f)) =
      "not (" ^ create(f) ^ ")"
  | create(Forall(v, t, f)) =
      "(" ^ create(f) ^ ")"
  | create(Exists(v, t, f)) =
      raise Not_Useful_Test
  | create(Atomic(a)) =
      createA(a)
      
and createA(Equals(e0, e1)) =
      createE(e0) ^ " = " ^ createE(e1)
  | createA(GreaterThan(e0, e1)) =
      createE(e0) ^ " > " ^ createE(e1)
  | createA(LessThan(e0, e1)) =
      createE(e0) ^ " < " ^ createE(e1)
  | createA(Expr(e)) =
      createE(e)

and createE(Var(s)) = s
  | createE(Vars(list)) = displayList(list)
  | createE(Bool(b))    = Bool.toString(b)
  | createE(Int(i))     = Int.toString(i)  
  | createE(Stringg(s)) = "\"" ^ s ^ "\""
  | createE(Charr(c))   = "#\"" ^ c ^ "\""
  | createE(Op(s, e0, e1)) =
      "(" ^ createE(e0) ^ " " ^ s ^ " " ^ createE(e1) ^ ")"
  | createE(App(e0, e1)) =
      createE(e0) ^ "(" ^ createE(e1) ^ ")"
  | createE(Andalso(e0, e1)) =
      createE(e0) ^ " andalso " ^ createE(e1)
  | createE(Orelse(e0, e1)) =
      createE(e0) ^ " orelse " ^ createE(e1)

and displayList([h])  = createE(h)
  | displayList(h::t) = createE(h) ^ ", " ^ displayList(t)
  | displayList([])   = ""

and stripVars([(v, t)])     = v
  | stripVars((v, t)::vars) = v ^ ", " ^ stripVars(vars)
  | stripVars([])           = ""

and make(f) = if checkQuantifiers(f)
              then makeImplies(simpleTransform(f), getVars(f, []))
              else makeNoImplies(fullTransform(f), getVars(f, []))
              
and makeImplies(Forall(v, t, f), vars) = makeImplies(f, vars)
  | makeImplies(Implies(f0, f1), vars) =
      makeImplies2(f0, vars) ^ " implies " ^ makeImplies(f1, vars)
  | makeImplies(f, vars) =
      "pred(fn (" ^ stripVars(vars) ^ ") => " ^ create(f) ^ ")"

and makeImplies2(f, vars) =
      "(fn (" ^ stripVars(vars) ^ ") => " ^ create(f) ^ ")"
      
and makeNoImplies(f, vars) =
      "pred(fn (" ^ stripVars(vars) ^ ") => " ^ create(f) ^ ")";


(* If the tag is longer than 51 characters then this causes a problem
 * with the output display. We also need to check incase we need to
 * add any escape sequences.
 *)

fun trim(s) =
      let
        val chars = explode(s)
      in
        implode(crop(chars, 0))
      end

and crop(h::t, count) =
      if count < 46
      then if h = #"\""
           then [#"\\", #"\""]@crop(t, count+2)
           else [h]@crop(t, count+1)
      else [#".", #".", #"."]
  | crop([], count) = [];

(* Create the QCheck command using all the pieces defined above. *)

fun makeCommand(input) = let
                           val parsed = parse(input)
			 in
			   "checkGen (" ^ makeGen(getVars(parsed, [])) ^ 
				") (\"" ^ trim(input) ^ "\", " ^
				make(parsed) ^ ")"
                          end;


(* Run the test by creating a QCheck command, writing it to a file
 * then read it back in again to run it.
 *)

fun qecheck(input) = let
                       val outfile = openOut("temp.qec")
                       val command = makeCommand(input)
                     in
                       output(outfile, makeCommand(input));
                       flushOut(outfile);
                       use "temp.qec"
                     end;
